//======================================================================================================================
//
//  This file is part of waLBerla. waLBerla is free software: you can
//  redistribute it and/or modify it under the terms of the GNU General Public
//  License as published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
//  waLBerla is distributed in the hope that it will be useful, but WITHOUT
//  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
//  for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with waLBerla (see COPYING.txt). If not, see <http://www.gnu.org/licenses/>.
//
//! \file UniformGridGPU.cpp
//! \author Martin Bauer <martin.bauer@fau.de>
//! \author Frederik Hennig <frederik.hennig@fau.de>
//! \author Markus Holzer <markus.holzer@fau.de>
//
//======================================================================================================================

#include "blockforest/Initialization.h"

#include "core/Environment.h"
#include "core/MemoryUsage.h"
#include "core/logging/Initialization.h"
#include "core/timing/RemainingTimeLogger.h"
#include "core/timing/TimingPool.h"

#include "field/AddToStorage.h"
#include "field/FlagField.h"
#include "field/vtk/VTKWriter.h"

#include "geometry/InitBoundaryHandling.h"

#include "gpu/AddGPUFieldToStorage.h"
#include "gpu/DeviceSelectMPI.h"
#include "gpu/FieldCopy.h"
#include "gpu/GPUWrapper.h"
#include "gpu/HostFieldAllocator.h"
#include "gpu/ParallelStreams.h"
#include "gpu/communication/UniformGPUScheme.h"

#include "lbm_generated/field/PdfField.h"
#include "lbm_generated/field/AddToStorage.h"
#include "lbm_generated/gpu/UniformGeneratedGPUPdfPackInfo.h"
#include "lbm_generated/gpu/GPUPdfField.h"
#include "lbm_generated/gpu/AddToStorage.h"
#include "lbm_generated/evaluation/PerformanceEvaluation.h"

#include "python_coupling/CreateConfig.h"
#include "python_coupling/DictWrapper.h"
#include "python_coupling/PythonCallback.h"

#include "timeloop/SweepTimeloop.h"

#include <cmath>

#include "InitShearVelocity.h"
#include "UniformGridGPU_InfoHeader.h"

using namespace walberla;

using StorageSpecification_T = lbm::UniformGridGPUStorageSpecification;
using Stencil_T = lbm::UniformGridGPUStorageSpecification::Stencil;

using PdfField_T = lbm_generated::PdfField< StorageSpecification_T >;
using GPUPdfField_T = lbm_generated::GPUPdfField< StorageSpecification_T >;
using FlagField_T = FlagField< uint8_t >;
using BoundaryCollection_T = lbm::UniformGridGPUBoundaryCollection< FlagField_T >;

using SweepCollection_T = lbm::UniformGridGPUSweepCollection;

using gpu::communication::UniformGPUScheme;

using macroFieldType = VelocityField_T::value_type;

int main(int argc, char** argv)
{
   mpi::Environment const env(argc, argv);
   gpu::selectDeviceBasedOnMpiRank();

   for (auto cfg = python_coupling::configBegin(argc, argv); cfg != python_coupling::configEnd(); ++cfg)
   {
      WALBERLA_MPI_WORLD_BARRIER()

      WALBERLA_GPU_CHECK(gpuPeekAtLastError())

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      ///                                        SETUP AND CONFIGURATION                                             ///
      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

      auto config = *cfg;
      logging::configureLogging(config);
      auto blocks = blockforest::createUniformBlockGridFromConfig(config);

      // Reading parameters
      auto parameters          = config->getOneBlock("Parameters");
      const real_t omega       = parameters.getParameter< real_t >("omega", real_c(1.4));
      const uint_t timesteps   = parameters.getParameter< uint_t >("timesteps", uint_c(50));
      const bool initShearFlow = parameters.getParameter< bool >("initShearFlow", true);
      const bool cudaEnabledMPI = parameters.getParameter< bool >("cudaEnabledMPI", false);

      // Creating fields
      const StorageSpecification_T StorageSpec = StorageSpecification_T();
      const BlockDataID pdfFieldCpuID  = lbm_generated::addPdfFieldToStorage(blocks, "pdfs", StorageSpec, uint_c(1), field::fzyx);

      auto allocator = make_shared< gpu::HostFieldAllocator<macroFieldType> >(); // use pinned memory allocator for faster CPU-GPU memory transfers
      const BlockDataID velFieldCpuID = field::addToStorage< VelocityField_T >(blocks, "vel", macroFieldType(0.0), field::fzyx, uint_c(1), allocator);
      const BlockDataID densityFieldCpuID = field::addToStorage< ScalarField_T >(blocks, "density", macroFieldType(1.0), field::fzyx, uint_c(1), allocator);
      const BlockDataID flagFieldID = field::addFlagFieldToStorage< FlagField_T >(blocks, "Boundary Flag Field");

      // Initialize velocity on cpu
      if (initShearFlow)
      {
         WALBERLA_LOG_INFO_ON_ROOT("Initializing shear flow")
         initShearVelocity(blocks, velFieldCpuID);
      }

      const BlockDataID pdfFieldGpuID = lbm_generated::addGPUPdfFieldToStorage< PdfField_T >(blocks, pdfFieldCpuID, StorageSpec, "pdfs on GPU", true);
      const BlockDataID velFieldGpuID =
         gpu::addGPUFieldToStorage< VelocityField_T >(blocks, velFieldCpuID, "velocity on GPU", true);
      const BlockDataID densityFieldGpuID =
         gpu::addGPUFieldToStorage< ScalarField_T >(blocks, densityFieldCpuID, "velocity on GPU", true);

      const Cell innerOuterSplit = Cell(parameters.getParameter< Vector3<cell_idx_t> >("innerOuterSplit", Vector3<cell_idx_t>(1, 1, 1)));
      Vector3< int32_t > gpuBlockSize = parameters.getParameter< Vector3< int32_t > >("gpuBlockSize", Vector3< int32_t >(256, 1, 1));
      SweepCollection_T sweepCollection(blocks, pdfFieldGpuID, densityFieldGpuID, velFieldGpuID, gpuBlockSize[0], gpuBlockSize[1], gpuBlockSize[2], omega, innerOuterSplit);
      for (auto& block : *blocks)
      {
         sweepCollection.initialise(&block);
      }

      int streamHighPriority = 0;
      int streamLowPriority  = 0;
      WALBERLA_GPU_CHECK(gpuDeviceGetStreamPriorityRange(&streamLowPriority, &streamHighPriority))
      sweepCollection.setOuterPriority(streamHighPriority);
      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      ///                                      LB SWEEPS AND BOUNDARY HANDLING                                       ///
      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      const pystencils::UniformGridGPU_StreamOnlyKernel StreamOnlyKernel(pdfFieldGpuID);

      // Boundaries
      const FlagUID fluidFlagUID("Fluid");
      auto boundariesConfig   = config->getBlock("Boundaries");
      if (boundariesConfig)
      {
         WALBERLA_LOG_INFO_ON_ROOT("Setting boundary conditions")
         geometry::initBoundaryHandling< FlagField_T >(*blocks, flagFieldID, boundariesConfig);
      }
      geometry::setNonBoundaryCellsToDomain< FlagField_T >(*blocks, flagFieldID, fluidFlagUID);
      BoundaryCollection_T boundaryCollection(blocks, flagFieldID, pdfFieldGpuID, fluidFlagUID);

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      ///                                           COMMUNICATION SCHEME                                             ///
      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

      UniformGPUScheme< Stencil_T > communication(blocks, cudaEnabledMPI);
      auto packInfo = std::make_shared<lbm_generated::UniformGeneratedGPUPdfPackInfo< GPUPdfField_T >>(pdfFieldGpuID);
      communication.addPackInfo(packInfo);

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      ///                                          TIME STEP DEFINITIONS                                             ///
      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      SweepTimeloop timeLoop(blocks->getBlockStorage(), timesteps);
      const std::string timeStepStrategy = parameters.getParameter< std::string >("timeStepStrategy", "normal");

      auto defaultStream = gpu::StreamRAII::newPriorityStream(streamLowPriority);

      if (timeStepStrategy == "noOverlap") {
         if (boundariesConfig){
            timeLoop.add() << BeforeFunction(communication.getCommunicateFunctor(), "communication")
                           << Sweep(boundaryCollection.getSweep(BoundaryCollection_T::ALL, defaultStream), "Boundary Conditions");
            timeLoop.add() << Sweep(sweepCollection.streamCollide(SweepCollection_T::ALL, defaultStream), "LBM StreamCollide");
         }else {
            timeLoop.add() << BeforeFunction(communication.getCommunicateFunctor(), "communication")
                           << Sweep(sweepCollection.streamCollide(SweepCollection_T::ALL, defaultStream), "LBM StreamCollide");}

      } else if (timeStepStrategy == "simpleOverlap") {
         if (boundariesConfig){
            timeLoop.add() << BeforeFunction(communication.getStartCommunicateFunctor(), "Start Communication")
                           << Sweep(boundaryCollection.getSweep(BoundaryCollection_T::ALL, defaultStream), "Boundary Conditions");
            timeLoop.add() << Sweep(sweepCollection.streamCollide(SweepCollection_T::INNER, defaultStream), "LBM StreamCollide Inner Frame");
            timeLoop.add() << BeforeFunction(communication.getWaitFunctor(), "Wait for Communication")
                           << Sweep(sweepCollection.streamCollide(SweepCollection_T::OUTER, defaultStream), "LBM StreamCollide Outer Frame");
         }else{
            timeLoop.add() << BeforeFunction(communication.getStartCommunicateFunctor(), "Start Communication")
                           << Sweep(sweepCollection.streamCollide(SweepCollection_T::INNER, defaultStream), "LBM StreamCollide Inner Frame");
            timeLoop.add() << BeforeFunction(communication.getWaitFunctor(), "Wait for Communication")
                           << Sweep(sweepCollection.streamCollide(SweepCollection_T::OUTER,defaultStream), "LBM StreamCollide Outer Frame");}

      } else if (timeStepStrategy == "kernelOnly") {
         timeLoop.add() << Sweep(sweepCollection.streamCollide(SweepCollection_T::ALL, defaultStream), "LBM StreamCollide");
      } else if (timeStepStrategy == "StreamOnly") {
         timeLoop.add() << Sweep(StreamOnlyKernel, "LBM Stream Only");
      } else {
         WALBERLA_ABORT_NO_DEBUG_INFO("Invalid value for 'timeStepStrategy'")
      }

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      ///                                             TIME LOOP SETUP                                                ///
      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

      // VTK
      const uint_t vtkWriteFrequency = parameters.getParameter< uint_t >("vtkWriteFrequency", 0);
      if (vtkWriteFrequency > 0)
      {
         auto vtkOutput = vtk::createVTKOutput_BlockData(*blocks, "vtk", vtkWriteFrequency, 0, false, "vtk_out",
                                                         "simulation_step", false, true, true, false, 0);
         auto velWriter = make_shared< field::VTKWriter< VelocityField_T, float32 > >(velFieldCpuID, "vel");
         vtkOutput->addCellDataWriter(velWriter);

         vtkOutput->addBeforeFunction([&]() {
            for (auto& block : *blocks)
               sweepCollection.calculateMacroscopicParameters(&block);
            gpu::fieldCpy< VelocityField_T, gpu::GPUField< VelocityField_T::value_type > >(blocks, velFieldCpuID, velFieldGpuID);
         });
         timeLoop.addFuncAfterTimeStep(vtk::writeFiles(vtkOutput), "VTK Output");
      }

      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      ///                                               BENCHMARK                                                    ///
      //////////////////////////////////////////////////////////////////////////////////////////////////////////////////

      lbm_generated::PerformanceEvaluation<FlagField_T> const performance(blocks, flagFieldID, fluidFlagUID);
      const uint_t warmupSteps     = parameters.getParameter< uint_t >("warmupSteps", uint_c(2));
      const uint_t outerIterations = parameters.getParameter< uint_t >("outerIterations", uint_c(1));
      for (uint_t i = 0; i < warmupSteps; ++i)
         timeLoop.singleStep();

      auto remainingTimeLoggerFrequency =
         parameters.getParameter< real_t >("remainingTimeLoggerFrequency", real_c(-1.0)); // in seconds
      if (remainingTimeLoggerFrequency > 0)
      {
         auto logger = timing::RemainingTimeLogger(timeLoop.getNrOfTimeSteps() * uint_c(outerIterations),
                                                   remainingTimeLoggerFrequency);
         timeLoop.addFuncAfterTimeStep(logger, "remaining time logger");
      }

      WcTimingPool timeloopTiming;
      // FIXME: I guess these two must be initialized first before I can call merge on them...
      WcTimer timerCombiningOuterAndThreadMeasurements; //= {0, 0.0, 0.0, 0.0, 0.0};
      WcTimer timeToSolution;
      for (uint_t outerIteration = 1; outerIteration <= outerIterations; ++outerIteration)
      {
         WALBERLA_GPU_CHECK(gpuPeekAtLastError())

         timeLoop.setCurrentTimeStepToZero();

         WALBERLA_GPU_CHECK( gpuDeviceSynchronize() )
         WALBERLA_GPU_CHECK( gpuPeekAtLastError() )
         WALBERLA_LOG_INFO_ON_ROOT("Starting outer-loop [" << outerIteration << "/" << outerIterations << "] for simulation with " << timesteps << " time steps")
         WcTimer simTimer;
         simTimer.reset();
         timeLoop.run(timeloopTiming);
         WALBERLA_GPU_CHECK( gpuDeviceSynchronize() )
         simTimer.end();
         WALBERLA_LOG_INFO_ON_ROOT("Simulation finished")
         
         timerCombiningOuterAndThreadMeasurements.merge(simTimer);
         // Get the runtime that the simulation takes for t timesteps to finish.
         double timeToSolutionForCurrentOuterIteration = walberla::mpi::reduce(simTimer.max(), walberla::mpi::MAX);
         timeToSolution.merge(WcTimer{/*counter, min, max, total, sumOfSquares*/
            1, 
            timeToSolutionForCurrentOuterIteration,
            timeToSolutionForCurrentOuterIteration,
            timeToSolutionForCurrentOuterIteration,
            timeToSolutionForCurrentOuterIteration*timeToSolutionForCurrentOuterIteration
         });
         
         // Prints timing statistics for every outer iteration
         // TODO: to much information, should be a verbose option only.
         // performance.logResultOnRoot(timesteps, timeToSolutionForCurrentOuterIteration);
         WALBERLA_LOG_RESULT_ON_ROOT("Outer step [" << outerIteration << "/" << outerIterations << "] took " << timeToSolutionForCurrentOuterIteration << " seconds.")

      } // END outer loop

      const auto reducedTimeloopTiming = timeloopTiming.getReduced();
      WALBERLA_LOG_RESULT_ON_ROOT("Time loop timing:\n" << *reducedTimeloopTiming)
      printResidentMemoryStatistics();

      // NOTE: This timeToSolution timer is a statistic of `outerIterations` repetitions of the same entire simulation.
      //    The `simTimer` times an entire simulation, i.e. the specified number of time steps (`timesteps`).
      //    The MAX value of one `simTimer` is the time, the slowest process took for completen, i.e., the time felt by a user.
      //    The entire simulation is repeated `outerIterations` times to create a meaningful timing statistic.       
      //    -> Hence, only `average`, `min`, and `max` are meaningful metrics to describe the simulation runtime.
      WALBERLA_LOG_RESULT_ON_ROOT(
         "Timing statistics for " << timesteps << " time steps over " << outerIterations << " outer iterations:" 
         << "\n\tTotal   :\t" << timeToSolution.total()
         << "\n\tCount   :\t" << timeToSolution.getCounter()
         << "\n\tAverage :\t" << timeToSolution.average()
         << "\n\tVariance:\t" << timeToSolution.variance()
         << "\n\tMin     :\t" << timeToSolution.min()
         << "\n\tMax     :\t" << timeToSolution.max()
      )
      performance.logResultOnRoot(timesteps, timeToSolution.average());
     
      // FIXME: maybe remove this as it is too much information:
      auto advancedTimer = WcTimer(
         walberla::mpi::reduce(timerCombiningOuterAndThreadMeasurements.getCounter(), walberla::mpi::SUM),
         walberla::mpi::reduce(timerCombiningOuterAndThreadMeasurements.min(), walberla::mpi::MIN),
         walberla::mpi::reduce(timerCombiningOuterAndThreadMeasurements.max(), walberla::mpi::MAX),
         walberla::mpi::reduce(timerCombiningOuterAndThreadMeasurements.total(), walberla::mpi::SUM),
         walberla::mpi::reduce(timerCombiningOuterAndThreadMeasurements.sumOfSquares(), walberla::mpi::SUM)
      );
      WALBERLA_LOG_RESULT_ON_ROOT(
         "Advanced timing statistics for " << timesteps << " time steps over " << outerIterations << " outer iterations considering the timing of every " << performance.threads() << " processes individually." 
         << "\n\tTotal   :\t" << advancedTimer.total()
         << "\n\tCount   :\t" << advancedTimer.getCounter()
         << "\n\tAverage :\t" << advancedTimer.average()
         << "\n\tVariance:\t" << advancedTimer.variance()
         << "\n\tMin     :\t" << advancedTimer.min()
         << "\n\tMax     :\t" << advancedTimer.max()
      )

      WALBERLA_ROOT_SECTION()
      {
         python_coupling::PythonCallback pythonCallbackResults("results_callback");
         if (pythonCallbackResults.isCallable())
         {
            pythonCallbackResults.data().exposeValue("numProcesses", performance.processes());
            pythonCallbackResults.data().exposeValue("numThreads", performance.threads());
            pythonCallbackResults.data().exposeValue("numCores", performance.cores());
            pythonCallbackResults.data().exposeValue("numberOfCells", performance.numberOfCells());
            pythonCallbackResults.data().exposeValue("numberOfFluidCells", performance.numberOfFluidCells());
            pythonCallbackResults.data().exposeValue("mlups", performance.mlups(timesteps, timeToSolution.average()));
            pythonCallbackResults.data().exposeValue("mlupsPerCore", performance.mlupsPerCore(timesteps, timeToSolution.average()));
            pythonCallbackResults.data().exposeValue("mlupsPerProcess", performance.mlupsPerProcess(timesteps, timeToSolution.average()));
            pythonCallbackResults.data().exposeValue("stencil", infoStencil);
            pythonCallbackResults.data().exposeValue("streamingPattern", infoStreamingPattern);
            pythonCallbackResults.data().exposeValue("collisionSetup", infoCollisionSetup);
            pythonCallbackResults.data().exposeValue("cse_global", infoCseGlobal);
            pythonCallbackResults.data().exposeValue("counter_for_time", timeToSolution.getCounter());
            pythonCallbackResults.data().exposeValue("total_time", timeToSolution.total());
            pythonCallbackResults.data().exposeValue("average_time", timeToSolution.average());
            pythonCallbackResults.data().exposeValue("variance_time", timeToSolution.variance());
            pythonCallbackResults.data().exposeValue("min_time", timeToSolution.min());
            pythonCallbackResults.data().exposeValue("max_time", timeToSolution.max());
            pythonCallbackResults.data().exposeValue("combined_counter_for_time", advancedTimer.getCounter());
            pythonCallbackResults.data().exposeValue("combined_total_time", advancedTimer.total());
            pythonCallbackResults.data().exposeValue("combined_average_time", advancedTimer.average());
            pythonCallbackResults.data().exposeValue("combined_variance_time", advancedTimer.variance());
            pythonCallbackResults.data().exposeValue("combined_min_time", advancedTimer.min());
            pythonCallbackResults.data().exposeValue("combined_max_time", advancedTimer.max());
            pythonCallbackResults.data().exposeValue("timesteps", timesteps);
            pythonCallbackResults.data().exposeValue("timeStepsPerSecond", performance.timeStepsPerSecond(timesteps, timeToSolution.average()));
            pythonCallbackResults.data().exposeValue("fineTimeStepsPerSecond", performance.fineTimeStepsPerSecond(timesteps, timeToSolution.average()));
            pythonCallbackResults.data().exposeValue("combined_timeStepsPerSecond", performance.timeStepsPerSecond(timesteps, advancedTimer.average()));
            pythonCallbackResults.data().exposeValue("combined_fineTimeStepsPerSecond", performance.fineTimeStepsPerSecond(timesteps, advancedTimer.average()));


            auto stripWhiteSpaces = [](std::string str) {
               str.erase(remove_if(str.begin(), str.end(), isspace), str.end());
               return str;
            };

            for (auto i = reducedTimeloopTiming->begin(); i != reducedTimeloopTiming->end(); ++i)
            {
               const std::string timerName = "Timer_" + i->first;
               std::string columnName = timerName + "_Total";
               pythonCallbackResults.data().exposeValue(stripWhiteSpaces(columnName).c_str(), i->second.total());
               columnName = timerName + "_Average";
               pythonCallbackResults.data().exposeValue(stripWhiteSpaces(columnName).c_str(), i->second.average());
               columnName = timerName + "_Count";
               pythonCallbackResults.data().exposeValue(stripWhiteSpaces(columnName).c_str(), i->second.getCounter());
               columnName = timerName + "_Min";
               pythonCallbackResults.data().exposeValue(stripWhiteSpaces(columnName).c_str(), i->second.min());
               columnName = timerName + "_Max";
               pythonCallbackResults.data().exposeValue(stripWhiteSpaces(columnName).c_str(), i->second.max());
               columnName = timerName + "_Variance";
               pythonCallbackResults.data().exposeValue(stripWhiteSpaces(columnName).c_str(), i->second.variance());
            }

            // Call Python function to report results
            pythonCallbackResults();
         }
      }
   }
   return EXIT_SUCCESS;
}
