import os
import waLBerla as wlb
from waLBerla.tools.config import decompose_into_3d, partition_domain
from waLBerla.tools.sqlitedb import sequenceValuesToScalars, checkAndUpdateSchema, storeSingle
import sys
import sqlite3
from math import prod

try:
    import machinestate as ms
except ImportError:
    ms = None

# Number of time steps run for a workload of 128^3 per GPU
# if double as many cells are on the GPU, half as many time steps are run etc.
# increase this to get more reliable measurements
TIME_STEPS_FOR_600_BLOCK = int(os.environ.get('TIME_STEPS_FOR_128_BLOCK', 37)) # Note:
    # The string above makes no sense.
    # For the sake of keeping the slum script functioning I left the name with 128_BLOCK.
DB_FILE = os.environ.get('DB_FILE', "gpu_benchmark.sqlite3")
BENCHMARK = int(os.environ.get('BENCHMARK', 0))

MAIN_MEMORY_GB = int(os.environ.get('MAIN_MEMORY_GB', 256)) # Default for MareNostrum5-GPP
GPU_MEMORY_GB = int(os.environ.get('GPU_MEMORY_GB', 64)) # Default for MareNostrum5-ACC
NUM_CORES_PER_NODE = int(os.environ.get('NUM_CORES_PER_NODE', 112)) # Default for MareNostrum5-GPP
NUM_GPUS_PER_NODE = int(os.environ.get('NUM_GPUS_PER_NODE', 4)) # Default for MareNostrum5-ACC
MAX_AVAIL_NODES = int(os.environ.get('MAX_AVAIL_NODES', 100)) # Default for MareNostrum5-GPP

WeakX = int(os.environ.get('WeakX', 600))
WeakY = int(os.environ.get('WeakY', 600))
WeakZ = int(os.environ.get('WeakZ', 600))

StrongX = int(os.environ.get('StrongX', 600))
StrongY = int(os.environ.get('StrongY', 600))
StrongZ = int(os.environ.get('StrongZ', 600))

BASE_CONFIG = {
    'DomainSetup': {
        'cellsPerBlock': (256, 128, 128),
        'periodic': (1, 1, 1),
    },
    'Parameters': {
        'omega': 1.8,
        'cudaEnabledMPI': False,
        'warmupSteps': 5,
        'outerIterations': 3,
    }
}

ldc_setup = {'Border': [
    {'direction': 'N', 'walldistance': -1, 'flag': 'UBB'},
    {'direction': 'W', 'walldistance': -1, 'flag': 'NoSlip'},
    {'direction': 'E', 'walldistance': -1, 'flag': 'NoSlip'},
    {'direction': 'S', 'walldistance': -1, 'flag': 'NoSlip'},
    {'direction': 'B', 'walldistance': -1, 'flag': 'NoSlip'},
    {'direction': 'T', 'walldistance': -1, 'flag': 'NoSlip'},
]}


def num_time_steps(block_size, time_steps_for_600_block=TIME_STEPS_FOR_600_BLOCK):
    """
    Calculate the number of time steps based on the block size.

    This function computes the number of time steps required for a given block size
    by scaling the time steps that could be executed on one process within one second 
    for a 600x600x600 cells_per_block to the given cells_per_block size.

    Parameters:
    block_size (tuple): A tuple of three integers representing the dimensions of the cells_per_block (x, y, z).
    time_steps_for_600_block (int, optional): The number of time steps for a 600x600x600 block. Default is 70000, 
    which is approximately the number of timesteps that were executed within one second on one entire MareNostrum5-acc node.

    Returns:
    int: The calculated number of time steps, with a minimum value of 10.
    """
    cells = block_size[0] * block_size[1] * block_size[2]
    time_steps = (600 ** 3 / cells) * time_steps_for_600_block
    if time_steps < 10:
        time_steps = 10
    return int(time_steps)


def cuda_block_size_ok(block_size, regs_per_threads=168):
    """
    Checks if a given CUDA block size does not exceed the SM register limit.
    168 registers per thread was obtained using cuobjdump on both SRT and Cumulant
    kernels. You might want to validate that for your own kernels.
    """
    return prod(block_size) * regs_per_threads < 64 * (2 ** 10) # The H100 has 64K 32-bit registers


def domain_block_size_ok(block_size, total_mem, gls=1, q=27, size_per_value=8):
    """
    Checks if a single block of given size fits into GPU memory.
    """
    return prod(b + 2 * gls for b in block_size) * q * size_per_value < total_mem


class Scenario:
    def __init__(self, benchmark_name="not defined", cells_per_block=(256, 128, 128), periodic=(1, 1, 1), cuda_blocks=(128, 1, 1),
                 timesteps=None, time_step_strategy="normal", omega=1.8, cuda_enabled_mpi=False,
                 inner_outer_split=(1, 1, 1), warmup_steps=5, outer_iterations=3,
                 init_shear_flow=False, boundary_setup=False,
                 vtk_write_frequency=0, remaining_time_logger_frequency=-1,
                 additional_info=None, blocks=None, db_file_name=None):

        self.benchmark_name = benchmark_name

        if boundary_setup:
            init_shear_flow = False
            periodic = (0, 0, 0)

        self.blocks = blocks if blocks else decompose_into_3d(wlb.mpi.numProcesses())

        self.cells_per_block = cells_per_block
        self.periodic = periodic

        self.time_step_strategy = time_step_strategy
        self.omega = omega
        self.timesteps = timesteps if timesteps else num_time_steps(cells_per_block)
        self.cuda_enabled_mpi = cuda_enabled_mpi
        self.inner_outer_split = inner_outer_split
        self.init_shear_flow = init_shear_flow
        self.boundary_setup = boundary_setup
        self.warmup_steps = warmup_steps
        self.outer_iterations = outer_iterations
        self.cuda_blocks = cuda_blocks

        self.vtk_write_frequency = vtk_write_frequency
        self.remaining_time_logger_frequency = remaining_time_logger_frequency
        self.db_file_name = DB_FILE if db_file_name is None else db_file_name

        self.config_dict = self.config(print_dict=False)
        self.additional_info = additional_info

    @wlb.member_callback
    def config(self, print_dict=True):
        from pprint import pformat
        config_dict = {
            'DomainSetup': {
                'blocks': self.blocks,
                'cellsPerBlock': self.cells_per_block,
                'periodic': self.periodic,
            },
            'Parameters': {
                'omega': self.omega,
                'cudaEnabledMPI': self.cuda_enabled_mpi,
                'warmupSteps': self.warmup_steps,
                'outerIterations': self.outer_iterations,
                'timeStepStrategy': self.time_step_strategy,
                'timesteps': self.timesteps,
                'initShearFlow': self.init_shear_flow,
                'gpuBlockSize': self.cuda_blocks,
                'innerOuterSplit': self.inner_outer_split,
                'vtkWriteFrequency': self.vtk_write_frequency,
                'remainingTimeLoggerFrequency': self.remaining_time_logger_frequency
            },
            'Logging': {
                'logLevel': 'info',  # info progress detail tracing
            }
        }
        if self.boundary_setup:
            config_dict["Boundaries"] = ldc_setup

        if print_dict:
            wlb.log_info_on_root("Scenario:\n" + pformat(config_dict))
            if self.additional_info:
                wlb.log_info_on_root("Additional Info:\n" + pformat(self.additional_info))
        return config_dict

    @wlb.member_callback
    def results_callback(self, **kwargs):
        data = {}
        data.update(self.config_dict['Parameters'])
        data.update(self.config_dict['DomainSetup'])
        data.update(kwargs)

        if self.additional_info is not None:
            data.update(self.additional_info)

        data['Identifier'] = self.benchmark_name
        data['executable'] = sys.argv[0]
        data['compile_flags'] = wlb.build_info.compiler_flags
        data['walberla_version'] = wlb.build_info.version
        data['build_machine'] = wlb.build_info.build_machine

        if ms:
            state = ms.MachineState(extended=False, anonymous=True)
            state.generate()  # generate subclasses
            state.update()  # read information
            data["MachineState"] = str(state.get())
        else:
            print("MachineState module is not available. MachineState was not saved")

        sequenceValuesToScalars(data)

        result = data
        sequenceValuesToScalars(result)
        num_tries = 4
        table_name = f"GPU_runs_{data['stencil']}_{data['streamingPattern']}_{data['collisionSetup']}_{prod(self.blocks)}nodes_{'x'.join(map(str, self.cells_per_block))}cellsPerBlock"
        table_name = table_name.replace("-", "_")
        table_name = table_name.replace(" ", "_")

        for num_try in range(num_tries):
            try:
                checkAndUpdateSchema(result, table_name, self.db_file_name, alter_table=True)
                storeSingle(result, table_name, self.db_file_name)
                break
            except sqlite3.OperationalError as e:
                wlb.log_warning(f"Sqlite DB writing failed: try {num_try + 1}/{num_tries}  {str(e)}")


# -------------------------------------- Functions trying different parameter sets -----------------------------------


def weak_scaling_overlap(cuda_enabled_mpi=False):
    wlb.log_info_on_root("Running scaling benchmark with communication hiding")
    wlb.log_info_on_root("")

    scenarios = wlb.ScenarioManager()
    cells_per_block = (WeakX, WeakY, WeakZ)

    # overlap
    for t in ["simpleOverlap"]:
        for split in [(1, 1, 1), (8, 8, 8), (16, 1, 1)]:
            if all(2 * s < c for s, c in zip(split, cells_per_block)):
                scenarios.add(Scenario(benchmark_name="weakScalingUniformGrid",
                                    cells_per_block=cells_per_block,
                                    cuda_blocks=(128, 1, 1),
                                    time_step_strategy=t,
                                    inner_outer_split=split,
                                    cuda_enabled_mpi=cuda_enabled_mpi,
                                    outer_iterations=2,
                                    timesteps=20,
                                    boundary_setup=True,
                                    db_file_name="weakScalingUniformGrid.sqlite3"))
            else:
                wlb.log_info_on_root(f"Split {split} would exceed cells per block {cells_per_block}. Skipping.")


def strong_scaling_overlap(cuda_enabled_mpi=False):
    wlb.log_info_on_root("Running strong scaling benchmark with one block per proc with communication hiding")
    wlb.log_info_on_root("")

    scenarios = wlb.ScenarioManager()

    total_domain_size = prod([StrongX, StrongY, StrongZ])
    num_processes = wlb.mpi.numProcesses()
    blocks = decompose_into_3d(num_processes)
    outer_iterations = max(1, int(0.85 * num_processes / NUM_GPUS_PER_NODE)) # C * #nodes
    # |-> C=0.85: `slowdown factor` to ensure that for many nodes, the run does not get too long.

    cells_per_block = tuple(partition_domain(total_domain_size, *reversed(blocks)))

    for t in ["simpleOverlap"]:
        for split in [(1, 1, 1), (8, 8, 8), (16, 1, 1)]:
            if all(2 * s < c for s, c in zip(split, cells_per_block)):
                scenarios.add(Scenario(benchmark_name="strongScalingUniformGridOneBlock",
                                    cells_per_block=cells_per_block,
                                    cuda_blocks=(128, 1, 1),
                                    time_step_strategy=t,
                                    inner_outer_split=split,
                                    cuda_enabled_mpi=cuda_enabled_mpi,
                                    outer_iterations=2*outer_iterations,
                                    timesteps=20,
                                    blocks=blocks,
                                    boundary_setup=True,
                                    db_file_name="strongScalingUniformGridOneBlock.sqlite3"))
            else:
                wlb.log_info_on_root(f"Split {split} would exceed cells per block {cells_per_block}. Skipping.")


def single_node_benchmark():
    """Benchmarks only the LBM compute kernel"""
    wlb.log_info_on_root("Running single GPU benchmarks")
    wlb.log_info_on_root("")

    try:
        gpu_mem_gb = int(os.environ.get('GPU_MEMORY_GB'))
        gpu_mem = gpu_mem_gb * (2 ** 30) # GPU Main Memory * the factor for GiB
    except TypeError:
        try:
            gpu_mem_mb = int(os.environ.get('GPU_MEMORY_MB'))
            gpu_mem = gpu_mem_mb * (2 ** 20) # GPU Main Memory * the factor for MiB
        except TypeError:
            try:
                gpu_mem_kb = int(os.environ.get('GPU_MEMORY_KB'))
                gpu_mem = gpu_mem_kb * (2 ** 10) # GPU Main Memory * the factor for KiB
            except TypeError:
                wlb.log_info_on_root("The size of device memory is unknown, so runtime errors can occure do to GPU out of memory.")
                gpu_mem = GPU_MEMORY_GB * (2 ** 30) # GPU Main Memory * the factor for GiB
    gpu_type = os.environ.get('GPU_TYPE')

    additional_info = {}
    if gpu_type is not None:
        additional_info['gpu_type'] = gpu_type

    scenarios = wlb.ScenarioManager()
    domain_block_sizes = [(i, i, i) for i in (128, 256, 512, 600, 650)]
    cuda_blocks = [(128, 1, 1), (32, 4, 1), (16, 4, 2), (8, 4, 4)]
    for block_size in domain_block_sizes:
        for cuda_block_size in cuda_blocks:
            for split in [(1, 1, 1), (8, 8, 8), (16, 1, 1)]:
                # cuda_block_size = (256, 1, 1) and block_size = (64, 64, 64) would be cut to cuda_block_size = (64, 1, 1)
                if cuda_block_size > block_size:
                    continue
                if not cuda_block_size_ok(cuda_block_size):
                    wlb.log_info_on_root(f"Cuda block size {cuda_block_size} would exceed register limit. Skipping.")
                    continue
                if not domain_block_size_ok(block_size, gpu_mem):
                    wlb.log_info_on_root(f"Block size {block_size} would exceed GPU memory. Skipping.")
                    continue
                if all(2 * s < c for s, c in zip(split, block_size)):
                    scenario = Scenario(benchmark_name="singleNodeUniformGridOneBlock",
                                        cells_per_block=block_size,
                                        cuda_blocks=cuda_block_size,
                                        time_step_strategy='kernelOnly',
                                        inner_outer_split=split,
                                        timesteps=num_time_steps(block_size),
                                        outer_iterations=2,
                                        additional_info=additional_info)
                    scenarios.add(scenario)
                else:
                    wlb.log_info_on_root(f"Split {split} would exceed cells per block {block_size}. Skipping.")


# TODO add a single GPU benchmark!


def validation_run():
    """Run with full periodic shear flow or boundary scenario (ldc) to check if the code works"""
    wlb.log_info_on_root("Validation run")
    wlb.log_info_on_root("")

    time_step_strategy = "noOverlap"  # "simpleOverlap"

    scenarios = wlb.ScenarioManager()
    scenario = Scenario(benchmark_name="validationUniformGrid",
                        cells_per_block=(128, 128, 128),
                        time_step_strategy=time_step_strategy,
                        timesteps=15,
                        outer_iterations=5,
                        warmup_steps=5,
                        init_shear_flow=False,
                        boundary_setup=True,
                        vtk_write_frequency=5000,
                        remaining_time_logger_frequency=30)
    scenarios.add(scenario)


wlb.log_info_on_root(f"Batch run of benchmark scenarios, saving result to {DB_FILE}")
# Select the benchmark you want to run
# single_gpu_benchmark()  # benchmarks different CUDA block sizes and domain sizes and measures single GPU
# performance of compute kernel (no communication)
# overlap_benchmark()  # benchmarks different communication overlap options
# profiling()  # run only two timesteps on a smaller domain for profiling only
# validation_run()

if BENCHMARK == 0:
    single_node_benchmark()
elif BENCHMARK == 1:
    weak_scaling_overlap(True)
elif BENCHMARK == 2:
    strong_scaling_overlap(True)
else:
    validation_run()

