from itertools import combinations_with_replacement
from math import floor, ceil, log2, prod
from typing import Tuple, List
import numpy as np

# FIXME: Probably discard this method as it is not helpful to find problem sizes unequal the initial one.
def partition_domain_approximate_N(N, Dx, Dy, Dz, verbose: bool = False) -> Tuple[int, int, int]:
    """
    Partitions N data points among a 3D domain split into Dx, Dy, Dz subdomains.

    Args:
        N (int): Total number of data points.
        Dx (int): Number of subdomains in x-direction.
        Dy (int): Number of subdomains in y-direction.
        Dz (int): Number of subdomains in z-direction.

    Returns:
        (int, int, int): Number of data points per subdomain in (nx, ny, nz).

    NOTE: First draft and comments were generated using ChatGBT. Code was improved by hand and with AI interaction afterwards.
    """
    D = np.array([Dx, Dy, Dz])
    # Compute the total number of subdomains
    D_total = np.prod(D)
    D_sum = np.sum(D)

    # Compute the ideal number of points per subdomain
    N_sub = N / D_total  # This may not be an integer

    # Compute approximate cube root distribution
    cube_root = (N_sub) ** (1/3)
    n = np.array([max(1, round(cube_root * (d / D_sum))) for d in D])

    # Adjust for rounding errors to ensure total points remain close to N
    total_points = np.prod(n * D)

    if verbose:
        print(f"ESTIMATE: I want to have {N} points, but I get {total_points} points. So a difference of {N-total_points} points.")

    while total_points > N:
        # Reduce the largest dimension slightly to fit within N
        max_index = np.argmax(n)
        n[max_index] -= 1
        total_points = np.prod(n * D)

    if verbose:
        print(f"REDUCE: I want to have {N} points, but I get {total_points} points. So a difference of {N-total_points} points.")

    stay = True
    while stay and total_points < N:
        # Increase a dimension to reach N
        sorted_indices = np.argsort(n)
        condition0 = D_total * n[sorted_indices[1]] * n[sorted_indices[2]] * (n[sorted_indices[0]] + 1)
        condition1 = D_total * n[sorted_indices[0]] * n[sorted_indices[2]] * (n[sorted_indices[1]] + 1)
        condition2 = D_total * n[sorted_indices[0]] * n[sorted_indices[1]] * (n[sorted_indices[2]] + 1)
        if condition0 <= N:
            n[sorted_indices[0]] += 1
        elif condition1 <= N:
            n[sorted_indices[1]] += 1
        elif condition2 <= N:
            n[sorted_indices[2]] += 1
        else:
            stay = False
        total_points = np.prod(n * D)

    if verbose:
        print(f"FILL-UP: I want to have {N} points, but I get {total_points} points. So a difference of {N-total_points} points.")

    if N-total_points != 0:
        print(f"The domain could not ideally be decomposed.\nThere are {N-total_points} less then requested.")

    n = tuple(sorted(n, reverse=True)) #descending order

    return int(n[0]), int(n[1]), int(n[2])


def suggest_scaling_appropriate_sizes_around_N(N: int, search_range: int = 2) -> Tuple[int]:
    """
    Provides a set of numbers around target number `N` that are especially suited for scaling runs, in descending order.
    For scaling, it is favorable multiples of powers of two, to bisect your problem as often as possible.
    Numbers below `N` are preferred.

    Method:
    - Find the closest power of 2 that is smaller than `N`.
    - Search range of appropriate candidates is increased by starting with the a power of two that is `search_range` smaller than the closest power of 2.
    - The suggested numbers are all multiples of the starting power of 2.
    - Suggested multiples range from C in {1, ..., 2^(search_range+1)-1}, i.e. up to the next power of two that is larger than `N`.
    - The candidates are ordered in descending order.

    Parameters:
    N (int): Target number.
    search_range (int): The larger `search_range`, the further will the returned numbers deviate from `N`.

    Returns:
    Tuple[int]: List of candidate problem sizes.
    """
    first_candidate = 2**max(1, floor(log2(N)) - search_range)
    candidates = [first_candidate * factor for factor in range(1, 2**(search_range+1))]


    return tuple(reversed(candidates))


def suggest_best_3d_problem_decompositions_for_initial_single_node_performance_test(
        max_number_of_nodes: int,
        processes_per_node: int,
        memory: float,
        precision: int,
        size_q: int,
        safety_factor: float,
        name: str,
        prioritize_balanced_decompositions: bool = False,
        # TODO add a parameter to specify the number of candidates to return
        debug: bool = False,
) -> Tuple[List[Tuple[int, int, int]], List[int]]:
    """
    Suggests a set of 3D problem decompositions that are especially suited for single node performance tests.
    The suggested decompositions are ordered in descending order of their total number of cells.

    Method:
    - The initial problem size is chosen to be the largest possible cube that fits into the available memory.
    - The cube root of the initial problem size is computed.
    - The cube root is used to suggest a set of problem sizes that are especially suited for scaling runs.
    - The suggested problem sizes are ordered in descending order of their total number of cells.

    Parameters:
    prioritize_balanced_decompositions (bool): If True, the candidates are ordered with respect to their balance. Otherwise, the candidates are ordered with respect to their total number of cells.
    debug (bool): If True, additional debug information is printed.

    Returns:
    List[List[Tuple[int, int, int]], List[int]]:
        List of suggested initial problem decompositions and
        a list containing the indices of decompositions in the first list, that allow to allocate the maximal number of nodes, even if they are not a power of two.
    """
    # Compute (N_max) the maximal number of cells that can be allocated on the given memory for given LBM model.
    N_max: int = floor(memory / safety_factor / precision / size_q)

    # Compute (n_max) the maximal number of cells that can be allocated per process (block) on one node,
    #   i.e., initial configuration strong scaling, or weak scaling.
    n_max: int = floor(N_max / processes_per_node)

    #   - compute the cube root
    n_j_max = round(n_max ** (1 / 3))
    #   - look for division friendly factors (in both directions) around the cube root
    n_j_candidates = suggest_scaling_appropriate_sizes_around_N(n_j_max, search_range=2)
    #   - create all possible combinations of candidates, i.e., the number of combinations is binomial_coefficient(k=3, n=len(n_j_candidates))
    decomposition_candidates = list(combinations_with_replacement(n_j_candidates, 3))
    #   - check which candidate combinations suffice the memory limit
    valid_n_j_candidates = list(filter(lambda num, n_max=n_max: np.prod(num) <= n_max, decomposition_candidates))
    #   - discard all but the first 10 combinations which product is closest to N_max
    largest_M_n_j_candidates = np.array(sorted(valid_n_j_candidates, key=prod, reverse=True)[0:10]) # FIXME I'm not sure if it is good to restrict here to 10 candidates?
    #   - check if I can bisect my problem until P_max
    list_of_scaling_processes = processes_per_node * 2**np.arange(ceil(log2(max_number_of_nodes+1)))
    condition_to_be_used_for_scaling = lambda N: all(N % process == 0 for process in list_of_scaling_processes)
    valid_candidates_for_scaling = [candidate for candidate in largest_M_n_j_candidates if condition_to_be_used_for_scaling(prod(candidate))]
    #   - sort them to be as equally distributed as possible
    reordered_candidates_for_scaling = sorted(valid_candidates_for_scaling, key=lambda dims: max(dims) - min(dims))
    #   - denote if P_max can be included or not
    index_of_scaling_candidates_that_allow_to_include_P_max = np.where(np.mod(np.prod(valid_candidates_for_scaling, axis=1), processes_per_node*max_number_of_nodes) == 0)[0].ravel()
    index_of_scaling_candidates_that_allow_to_include_P_max_reordered = np.where(np.mod(np.prod(reordered_candidates_for_scaling, axis=1), processes_per_node*max_number_of_nodes) == 0)[0].ravel()


    # DEBUG block
    if debug:
        relation_factor_n = np.max(np.array(valid_candidates_for_scaling)[0])
        utilization = (processes_per_node * np.prod(np.array(valid_candidates_for_scaling), axis=1))/N_max*100
        utilization_reordered = (processes_per_node * np.prod(np.array(reordered_candidates_for_scaling), axis=1))/N_max*100
        unbalance_factor = np.maximum(1, np.array(list(map(lambda dims: max(dims) - min(dims), valid_candidates_for_scaling)))/relation_factor_n*100)
        unbalance_factor_reordered = np.maximum(1, np.array(list(map(lambda dims: max(dims) - min(dims), reordered_candidates_for_scaling)))/relation_factor_n*100)
        size_order_score = utilization / unbalance_factor
        balance_order_score = utilization_reordered / unbalance_factor_reordered
        print("\n", 50*"-", "\nInformation for best suited candidate on ", name)
        print("Utilization      [without reordering] is ", utilization, " %."                              )
        print("Utilization      [with    reordering] is ", utilization_reordered, " %."            )
        print("Unbalance factor [without reordering] is ", unbalance_factor, " %."                    )
        print("Unbalance factor [with    reordering] is ", unbalance_factor_reordered, " %.")
        print("Total score      [with    reordering] is ", size_order_score, "."                      )
        print("Total score      [without reordering] is ", balance_order_score, "."                )
        print("All available nodes can be included for candidates of index [without reordering]: ", index_of_scaling_candidates_that_allow_to_include_P_max)
        print("All available nodes can be included for candidates of index [with    reordering]: ", index_of_scaling_candidates_that_allow_to_include_P_max_reordered)
        print(valid_candidates_for_scaling)
        print(reordered_candidates_for_scaling)
        print("Use 3-5 of these candidates for single node performance tests. Then use the best one for scaling.")

        # TODO change all list operations to numpy ones
        # TODO create functions out of the sub-steps

    if prioritize_balanced_decompositions:
        return reordered_candidates_for_scaling, index_of_scaling_candidates_that_allow_to_include_P_max_reordered
    else:
        return valid_candidates_for_scaling, index_of_scaling_candidates_that_allow_to_include_P_max



if __name__ == "__main__":
    # Example usage
    for max_number_of_nodes, processes_per_node, memory, precision, size_q, safety_factor, name in [
        ( 100, 112, 2.56e11, 8, 27, 1.4, "MareNostrum5-GPP"),
        ( 100,   1, 6.40e10, 8, 19, 1.2, "MareNostrum5-ACC"), # NOTE: you have 4 times the GPUs but each one has it's own memory
        ( 512, 128, 2.56e11, 8, 27, 1.4, "LUMI-C"),
        ( 128, 128, 5.12e11, 8, 27, 1.4, "LUMI-C-big"),
        (1024,   1, 1.28e11, 8, 19, 1.2, "LUMI-G"), # NOTE: you have 8 times the GPUs but each one has it's own memory
    ]:
        _, _ = suggest_best_3d_problem_decompositions_for_initial_single_node_performance_test(
            max_number_of_nodes, processes_per_node, memory, precision, size_q, safety_factor, name, debug=True
        )
