from itertools import combinations_with_replacement
from enum import Enum
from math import floor
from typing import Tuple, List

import numpy as np

class Condition(Enum):
    LESS_THAN = "<"
    GREATER_THAN = ">"
    EQUAL = "=="
    NOT_EQUAL = "!="
    LESS_EQUAL = "<="
    GREATER_EQUAL = ">="
    ALLOW_ALL = "allow_all"

def _factors(n):
    while n > 1:
        for i in range(2, n + 1):
            if n % i == 0:
                n //= i
                yield i
                break


def block_decomposition(processes):
    """Compute a 3D factorization of 'processes' in a 3 factors returned as tuple"""
    result = [1, 1, 1]
    for factor in _factors(processes):
        min_idx = result.index(min(result))
        result[min_idx] *= factor
    assert result[0] * result[1] * result[2] == processes
    return tuple(result)


def approximate_3d_factorizations_for_N(
    N: int,
    condition: Condition = Condition.EQUAL,
    search_range: int = 1,
    descending: bool = False,
) -> List[Tuple[int, int, int]]:
    """
    Factorize a given number 'N' into three dimensions (Dx, Dy, Dz) such that:
    - Dx * Dy * Dz satisfies the given condition with N (e.g. `==`, `<=`, `>=`, `!=`, `<`, `>`)
    - The decompositions prioritize the most balanced ones.

    Parameters:
        N (int): The target product value.
        condition (Condition): The condition to be a valid candidate.
        search_range (int): The range in which decomposed values vary around the cube root of N.
        descending (bool): If True, the factors for every factorization are stored in descending order (Dx ≥ Dy ≥ Dz), otherwise in ascending order (Dx ≤ Dy ≤ Dz).

    Returns:
        List[Tuple[int, int, int]]: All valid decompositions sorted by balance (smallest difference between min/max).
    """
    # Compute an approximate cube root of N
    cube_root = round(N ** (1 / 3))

    # Define the condition function
    def condition_func(_dx: int, _dy: int, _dz: int):
        if condition == Condition.LESS_THAN:
            return _dx * _dy * _dz < N
        elif condition == Condition.GREATER_THAN:
            return _dx * _dy * _dz > N
        elif condition == Condition.EQUAL:
            return _dx * _dy * _dz == N
        elif condition == Condition.NOT_EQUAL:
            return _dx * _dy * _dz != N
        elif condition == Condition.LESS_EQUAL:
            return _dx * _dy * _dz <= N
        elif condition == Condition.GREATER_EQUAL:
            return _dx * _dy * _dz >= N
        elif condition == Condition.ALLOW_ALL:
            return True

    # Generate ordered triplets where
    if descending:
        # Dx ≥ Dy ≥ Dz
        candidates = (
            (dx, dy, dz)
            for dx, dy, dz in combinations_with_replacement(
                reversed(range(max(1, cube_root - search_range), cube_root + search_range + 1)), 3
            )
            if condition_func(dx, dy, dz)
        )
    else:
        # Dx ≤ Dy ≤ Dz
        candidates = (
            (dx, dy, dz)
            for dx, dy, dz in combinations_with_replacement(
                range(max(1, cube_root - search_range), cube_root + search_range + 1), 3
            )
            if condition_func(dx, dy, dz)
        )

    # Sort results based on balance (minimizing max(dims) - min(dims))
    return sorted(candidates, key=lambda dims: max(dims) - min(dims))
    # NOTE: I'm not sure if it would not be better to keep the last two dimensions similar in size and the first smaller
    # return sorted(candidates, key=lambda dims: dims[2] - dims[1])

def get_best_3d_decomposition(N: int,
    descending: bool = False,
    debug: bool = False,
                              ) -> Tuple[int, int, int]:
    """
    Decomposes the number of N into three dimensions (Dx, Dy, Dz) such that:
    - Dx * Dy * Dz = N
    - Provides the most balanced decomposition (minimizing the distance between dimensions)
    - Repeats the search with exponentially increasing search range until a valid decomposition is found.

    Parameters:
        N (int): The total number of processes to decompose.
        descending (bool): Factors stored in descending order (Dx ≥ Dy ≥ Dz), otherwise in ascending order (Dx ≤ Dy ≤ Dz).

    Returns:
        Tuple[int, int, int]: A tuple representing the dimensions (Dx, Dy, Dz).
    """

    for search_range in 2**np.arange(0, floor(N ** (1/3))):
        options = approximate_3d_factorizations_for_N(N, condition=Condition.EQUAL, search_range=search_range, descending=descending)
        if debug:
            print(f"Search range: {search_range}, Options: {options}")
        if options:
            return options[0]
    return (1, 1, N)

if __name__ == "__main__":
    # Example usage
    for N in [100, 128, 512, 600, 1024, 1188, 2048, 1024*128, 2024*128]:
        print("\n", 50*"-", f"\nInformation for best suited candidate on N = {N}")
        dx, dy, dz = get_best_3d_decomposition(N, debug=True)
        print(f"Optimal decomposition: Dx = {dx}, Dy = {dy}, Dz = {dz}")