from math import log2, prod, floor

import numpy as np

from setup import get_best_3d_decomposition
from distributeProblemBetweenFixedDomain import suggest_best_3d_problem_decompositions_for_initial_single_node_performance_test

def scaling_dry_run(
        nx: int, ny: int, nz: int,
        include_max_nodes: bool,
        max_number_of_nodes: int,
        processes_per_node: int,
        memory: float,
        precision: int,
        size_q: int,
        safety_factor: float,
        name: str,
        ) -> None:
    N_max: int = floor(memory / safety_factor / precision / size_q)
    # TODO & FIXME don't decompose anew, just divide for blocks and cells the same dimension by two for every step.
    #   Only re-decompose for the max_number_of_nodes if they are not a power of two.
    #   Otherwise, the computation is to time-consuming for large numbers of nodes.
    N = prod([nx, ny, nz]) * processes_per_node
    print(50*"-")
    print(name, f"would fit {N_max:.3e} cells in total, {N:.3e} cells are used. ({100*N/N_max:.2f} %)")
    number_of_nodes_in_scaling_run = 2**np.arange(0, log2(max_number_of_nodes), dtype=np.uint32)
    if include_max_nodes:
        number_of_nodes_in_scaling_run = np.append(number_of_nodes_in_scaling_run, max_number_of_nodes)

    for nodes in number_of_nodes_in_scaling_run:
        assert(nodes.is_integer())
        processes = nodes * processes_per_node
        assert(processes.is_integer())
        n = N / processes
        assert(n.is_integer())

        blocks = get_best_3d_decomposition(processes)
        cells_per_block = get_best_3d_decomposition(n, descending=True)
        print(f"{nodes:4d} nodes: {blocks[0]:3d} x {blocks[1]:3d} x {blocks[2]:3d} blocks of {cells_per_block[0]:3d} x {cells_per_block[1]:3d} x {cells_per_block[2]:3d} cells each.")

    if not include_max_nodes:
        print(f"{max_number_of_nodes:3d} nodes: Are excluded for this configuration.")


if __name__ == "__main__":
    for max_number_of_nodes, processes_per_node, memory, precision, size_q, safety_factor, name in [
        ( 100, 112, 2.56e11, 8, 27, 1.4, "MareNostrum5-GPP"),
        ( 100,   1, 6.40e10, 8, 19, 1.2, "MareNostrum5-ACC"), # NOTE: you have 4 times the GPUs but each one has it's own memory
        ( 512, 128, 2.56e11, 8, 27, 1.4, "LUMI-C"),
        ( 128, 128, 5.12e11, 8, 27, 1.4, "LUMI-C-big"),
        (1024,   1, 1.28e11, 8, 19, 1.2, "LUMI-G"), # NOTE: you have 8 times the GPUs but each one has it's own memory
    ]:
        print("\n+", 20*"=", name, 20*"=", "+")
        dry_run_kernel = lambda nx, ny, nz, include_max_nodes: scaling_dry_run(nx, ny, nz, include_max_nodes, max_number_of_nodes, processes_per_node, memory, precision, size_q, safety_factor, name)

        problem_decomposition_candidates, large_run_indices = suggest_best_3d_problem_decompositions_for_initial_single_node_performance_test(
            max_number_of_nodes, processes_per_node, memory, precision, size_q, safety_factor, name
        )
        for index, [nx, ny, nz] in enumerate(problem_decomposition_candidates):
            if index in large_run_indices:
                print(f"\nSetup {index}: Suggested start decomposition is {nx} x {ny} x {nz} cells")
                dry_run_kernel(nx, ny, nz, True)
                break
        print("+", 50*"=", "+\n")
